@@ -43,7 +43,8 @@ html_static_path = ["./_static"]
 templates_path = ["./_templates"]
 html_css_files = [
     "custom.css",
-    "https://fonts.googleapis.com/css2?family=Fira+Code:wght@300;400;700&family=Fira+Sans:ital,wght@0,100;0,200;0,300;0,400;0,500;0,600;0,700;0,800;0,900;1,100;1,200;1,300;1,400;1,500;1,600;1,700;1,800;1,900&display=swap",
+    "colors.css",
+    "https://fonts.googleapis.com/css2?family=Fira+Code:wght@300;400;700&family=Fira+Sans:ital,wght@0,100;0,200;0,300;0,400;0,500;0,600;0,700;0,800;0,900;1,100;1,200;1,300;1,400;1,500;1,600;1,700;1,800;1,900&family=Be+Vietnam+Pro:wght@500&display=swap",
 ]
 
 html_theme_options.update({"light_logo": "tap-icon.png", "dark_logo": "tap-icon.png"})
@@ -78,7 +79,7 @@ autosectionlabel_maxdepth = 3
 autosectionlabel_prefix_document = True
 issues_github_path = "alisaifee/flask-limiter"
 
-ogp_image = "_static/logo.png"
+ogp_image = "_static/logo-og.png"
 
 extlinks = {
     "pypi": ("https://pypi.org/project/%s", "%s"),
