@@ -1,7 +1,13 @@
 from __future__ import annotations
 
 import dataclasses
+import traceback
 import warnings
+from types import TracebackType
+
+from ordered_set import OrderedSet
+
+from .util import get_qualified_name
 
 """
 Flask-Limiter Extension
@@ -13,7 +19,7 @@ import time
 import weakref
 from collections import defaultdict
 from functools import partial, wraps
-from typing import overload
+from typing import Type, overload
 
 import flask
 import flask.wrappers
@@ -45,15 +51,16 @@ from .wrappers import Limit, LimitGroup, RequestLimit
 
 @dataclasses.dataclass
 class LimiterContext:
-    rate_limiting_complete: bool = False
+    rate_limiting_complete: dict[str, bool] = dataclasses.field(default_factory=dict)
     view_rate_limit: Optional[RequestLimit] = None
     view_rate_limits: List[RequestLimit] = dataclasses.field(default_factory=list)
     conditional_deductions: Dict[Limit, List[str]] = dataclasses.field(
         default_factory=dict
     )
+    seen_limits: OrderedSet[Limit] = dataclasses.field(default_factory=OrderedSet)
 
     def reset(self) -> None:
-        self.rate_limiting_complete = False
+        self.rate_limiting_complete.clear()
         self.view_rate_limit = None
         self.view_rate_limits.clear()
         self.conditional_deductions.clear()
@@ -97,7 +104,7 @@ class Limiter:
      upon instantiation.
     :param auto_check: whether to automatically check the rate limit in
      the before_request chain of the application. default ``True``
-    :param swallow_errors: whether to swallow errors when hitting a rate
+    :param swallow_errors: whether to swallow any errors when hitting a rate
      limit. An exception will still be logged. default ``False``
     :param fail_on_first_breach: whether to stop processing remaining limits
      after the first breach. default ``True``
@@ -179,7 +186,6 @@ class Limiter:
         self._swallow_errors = swallow_errors
         self._fail_on_first_breach = fail_on_first_breach
         self._on_breach = on_breach
-
         # No longer optional
         assert key_func
 
@@ -189,17 +195,8 @@ class Limiter:
         _default_limits = (
             [
                 LimitGroup(
-                    limit,
-                    self._key_func,
-                    None,
-                    False,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    1,
+                    limit_provider=limit,
+                    key_function=self._key_func,
                 )
                 for limit in default_limits
             ]
@@ -210,17 +207,9 @@ class Limiter:
         _application_limits = (
             [
                 LimitGroup(
-                    limit,
-                    self._key_func,
-                    "global",
-                    False,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    1,
+                    limit_provider=limit,
+                    key_function=self._key_func,
+                    scope="global",
                 )
                 for limit in application_limits
             ]
@@ -232,17 +221,8 @@ class Limiter:
             for limit in in_memory_fallback:
                 self._in_memory_fallback.append(
                     LimitGroup(
-                        limit,
-                        self._key_func,
-                        None,
-                        False,
-                        None,
-                        None,
-                        None,
-                        None,
-                        None,
-                        None,
-                        1,
+                        limit_provider=limit,
+                        key_function=self._key_func,
                     )
                 )
 
@@ -260,8 +240,8 @@ class Limiter:
         self.limit_manager = LimitManager(
             application_limits=_application_limits,
             default_limits=_default_limits,
-            static_route_limits={},
-            dynamic_route_limits={},
+            static_decorated_limits={},
+            dynamic_decorated_limits={},
             static_blueprint_limits={},
             dynamic_blueprint_limits={},
             route_exemptions=self._route_exemptions,
@@ -372,17 +352,10 @@ class Limiter:
             self.limit_manager.set_application_limits(
                 [
                     LimitGroup(
-                        app_limits,
-                        self._key_func,
-                        "global",
-                        False,
-                        None,
-                        None,
-                        None,
-                        None,
-                        None,
-                        None,
-                        self._application_limits_cost,
+                        limit_provider=app_limits,
+                        key_function=self._key_func,
+                        scope="global",
+                        cost=self._application_limits_cost,
                     )
                 ]
             )
@@ -398,17 +371,12 @@ class Limiter:
             self.limit_manager.set_default_limits(
                 [
                     LimitGroup(
-                        conf_limits,
-                        self._key_func,
-                        None,
-                        self._default_limits_per_method,
-                        None,
-                        None,
-                        self._default_limits_exempt_when,
-                        None,
-                        self._default_limits_deduct_when,
-                        None,
-                        self._default_limits_cost,
+                        limit_provider=conf_limits,
+                        key_function=self._key_func,
+                        per_method=self._default_limits_per_method,
+                        exempt_when=self._default_limits_exempt_when,
+                        deduct_when=self._default_limits_deduct_when,
+                        cost=self._default_limits_cost,
                     )
                 ]
             )
@@ -461,7 +429,7 @@ class Limiter:
         cost: Union[int, Callable[[], int]] = 1,
     ) -> LimitDecorator:
         """
-        decorator to be used for rate limiting individual routes or blueprints.
+        Decorator to be used for rate limiting individual routes or blueprints.
 
         :param limit_value: rate limit string or a callable that returns a
          string. :ref:`ratelimit-string` for more details.
@@ -492,6 +460,17 @@ class Limiter:
          raised.
         :param cost: The cost of a hit or a function that
          takes no parameters and returns the cost as an integer (Default: ``1``).
+
+        Changes
+          - .. versionadded:: 2.9.0 The returned object can also be used as a context manager
+            for rate limiting a code block inside a view. For example::
+
+                @app.route("/")
+                def route():
+                   try:
+                       with limiter.limit("10/second"):
+                           # something expensive
+                   except RateLimitExceeded: pass
         """
 
         return LimitDecorator(
@@ -648,9 +627,7 @@ class Limiter:
         if isinstance(obj, flask.Blueprint):
             self.limit_manager.add_blueprint_exemption(obj.name, flags)
         elif obj:
-            self.limit_manager.add_route_exemption(
-                f"{obj.__module__}.{obj.__name__}", flags
-            )
+            self.limit_manager.add_route_exemption(get_qualified_name(obj), flags)
         else:
             _R = TypeVar("_R")
             _WO = TypeVar("_WO", Callable[..., _R], flask.Blueprint)
@@ -681,17 +658,11 @@ class Limiter:
         if not self._in_memory_fallback and fallback_limits:
             self._in_memory_fallback = [
                 LimitGroup(
-                    fallback_limits,
-                    self._key_func,
-                    None,
-                    False,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    None,
-                    1,
+                    limit_provider=fallback_limits,
+                    key_function=self._key_func,
+                    scope=None,
+                    per_method=False,
+                    cost=1,
                 )
             ]
 
@@ -725,7 +696,7 @@ class Limiter:
 
         :raises: RateLimitExceeded
         """
-        self._check_request_limit(False)
+        self._check_request_limit(in_middleware=False)
 
     def reset(self) -> None:
         """
@@ -875,28 +846,42 @@ class Limiter:
 
         return response
 
-    def __check_all_limits_exempt(self, endpoint: Optional[str]) -> bool:
+    def __check_all_limits_exempt(
+        self, endpoint: Optional[str], callable_name: Optional[str] = None
+    ) -> bool:
         return bool(
             not endpoint
             or not (self.enabled and self.initialized)
             or endpoint == "static"
             or any(fn() for fn in self._request_filters)
-            or self.context.rate_limiting_complete
+            or (
+                self.context.rate_limiting_complete.get(callable_name, False)
+                if callable_name
+                else any(self.context.rate_limiting_complete.values())
+            )
         )
 
     def __filter_limits(
         self,
         endpoint: Optional[str],
         blueprint: Optional[str],
+        callable_name: Optional[str],
         in_middleware: bool = False,
     ) -> List[Limit]:
-        view_func = flask.current_app.view_functions.get(endpoint or "", None)
-        name = f"{view_func.__module__}.{view_func.__name__}" if view_func else ""
 
-        if self.__check_all_limits_exempt(endpoint):
+        if callable_name:
+            name = callable_name
+        else:
+            view_func = flask.current_app.view_functions.get(endpoint or "", None)
+            name = get_qualified_name(view_func) if view_func else ""
+
+        if self.__check_all_limits_exempt(endpoint, callable_name):
             return []
 
-        marked_for_limiting = name in self._marked_for_limiting
+        marked_for_limiting = (
+            name in self._marked_for_limiting
+            or self.limit_manager.has_hints(endpoint or "")
+        )
         fallback_limits = []
 
         if self._storage_dead and self._fallback_limiter:
@@ -913,15 +898,18 @@ class Limiter:
                     self.__check_backend_count = 0
                 else:
                     fallback_limits = list(itertools.chain(*self._in_memory_fallback))
-
-        return self.limit_manager.resolve_limits(
+        resolved_limits = self.limit_manager.resolve_limits(
             flask.current_app,
             endpoint,
             blueprint,
+            name,
             in_middleware,
             marked_for_limiting,
             fallback_limits,
         )
+        limits = OrderedSet(resolved_limits) - self.context.seen_limits
+        self.context.seen_limits.update(limits)
+        return list(limits)
 
     def __evaluate_limits(self, endpoint: str, limits: List[Limit]) -> None:
         failed_limits: List[Tuple[Limit, List[str]]] = []
@@ -962,7 +950,7 @@ class Limiter:
             view_limits.append(RequestLimit(self, lim.limit, args, False))
 
             if not method(lim.limit, *args, **kwargs):
-                self.logger.warning(
+                self.logger.info(
                     "ratelimit %s (%s) exceeded at endpoint: %s",
                     lim.limit,
                     limit_key,
@@ -987,21 +975,28 @@ class Limiter:
                         if isinstance(cb_response, flask.wrappers.Response):
                             on_breach_response = cb_response
                     except Exception as err:  # noqa
-                        self.logger.warning(
-                            "on_breach callback failed with error %s", err
-                        )
-
+                        if self._swallow_errors:
+                            self.logger.exception(
+                                "on_breach callback failed with error %s", err
+                            )
+                        else:
+                            raise err
         if failed_limits:
             raise RateLimitExceeded(
                 sorted(failed_limits, key=lambda x: x[0].limit)[0][0],
                 response=on_breach_response,
             )
 
-    def _check_request_limit(self, in_middleware: bool = True) -> None:
+    def _check_request_limit(
+        self, callable_name: Optional[str] = None, in_middleware: bool = True
+    ) -> None:
         endpoint = flask.request.endpoint or ""
         try:
             all_limits = self.__filter_limits(
-                flask.request.endpoint, flask.request.blueprint, in_middleware
+                flask.request.endpoint,
+                flask.request.blueprint,
+                callable_name,
+                in_middleware,
             )
             self.__evaluate_limits(endpoint, all_limits)
         except Exception as e:
@@ -1014,7 +1009,10 @@ class Limiter:
                     " in-memory storage"
                 )
                 self._storage_dead = True
-                self._check_request_limit(in_middleware)
+                self.context.seen_limits.clear()
+                self._check_request_limit(
+                    callable_name=callable_name, in_middleware=in_middleware
+                )
             else:
                 if self._swallow_errors:
                     self.logger.exception("Failed to rate limit. Swallowing error")
@@ -1056,13 +1054,80 @@ class LimitDecorator:
         self.key_func = key_func or self.limiter._key_func
         self.scope = scope if shared else None
         self.per_method = per_method
-        self.methods = methods
+        self.methods = tuple(methods) if methods else None
         self.error_message = error_message
         self.exempt_when = exempt_when
         self.override_defaults = override_defaults
         self.deduct_when = deduct_when
         self.on_breach = on_breach
         self.cost = cost
+        self.is_static = not callable(self.limit_value)
+
+    @property
+    def dynamic_limit(self) -> Optional[LimitGroup]:
+        return LimitGroup(
+            limit_provider=self.limit_value,
+            key_function=self.key_func,
+            scope=self.scope,
+            per_method=self.per_method,
+            methods=self.methods,
+            error_message=self.error_message,
+            exempt_when=self.exempt_when,
+            override_defaults=self.override_defaults,
+            deduct_when=self.deduct_when,
+            on_breach=self.on_breach,
+            cost=self.cost,
+        )
+
+    @property
+    def static_limits(self) -> List[Limit]:
+        return list(
+            LimitGroup(
+                limit_provider=self.limit_value,
+                key_function=self.key_func,
+                scope=self.scope,
+                per_method=self.per_method,
+                methods=self.methods,
+                error_message=self.error_message,
+                exempt_when=self.exempt_when,
+                override_defaults=self.override_defaults,
+                deduct_when=self.deduct_when,
+                on_breach=self.on_breach,
+                cost=self.cost,
+            )
+        )
+
+    def __enter__(self) -> None:
+        tb = traceback.extract_stack(limit=2)
+        qualified_location = f"{tb[0].filename}:{tb[0].name}:{tb[0].lineno}"
+
+        # TODO: if use as a context manager becomes interesting/valuable
+        #  a less hacky approach than using the traceback and piggy backing
+        #  on the limit manager's knowledge of decorated limits might be worth it.
+        if not self.is_static:
+            self.limiter.limit_manager.add_decorated_runtime_limit(
+                qualified_location, self.dynamic_limit
+            )
+        else:
+            self.limiter.limit_manager.add_decorated_static_limit(
+                qualified_location, *self.static_limits
+            )
+
+        self.limiter.limit_manager.add_endpoint_hint(
+            flask.request.endpoint, qualified_location
+        )
+
+        self.limiter._check_request_limit(
+            in_middleware=False, callable_name=qualified_location
+        )
+
+    def __exit__(
+        self,
+        exc_type: Optional[Type[BaseException]],
+        exc_value: Optional[BaseException],
+        traceback: Optional[TracebackType],
+    ) -> None:
+        ...
 
     @overload
     def __call__(self, obj: Callable[P, R]) -> Callable[P, R]:
@@ -1079,41 +1144,12 @@ class LimitDecorator:
         if isinstance(obj, flask.Blueprint):
             name = obj.name
         else:
-            name = f"{obj.__module__}.{obj.__name__}"
-
-        dynamic_limit, static_limits = None, []
-
-        if callable(self.limit_value):
-            dynamic_limit = LimitGroup(
-                self.limit_value,
-                self.key_func,
-                self.scope,
-                self.per_method,
-                self.methods,
-                self.error_message,
-                self.exempt_when,
-                self.override_defaults,
-                self.deduct_when,
-                self.on_breach,
-                self.cost,
-            )
-        else:
+            name = get_qualified_name(obj)
+        dynamic_limit = self.dynamic_limit if not self.is_static else None
+        static_limits = []
+        if self.is_static:
             try:
-                static_limits = list(
-                    LimitGroup(
-                        self.limit_value,
-                        self.key_func,
-                        self.scope,
-                        self.per_method,
-                        self.methods,
-                        self.error_message,
-                        self.exempt_when,
-                        self.override_defaults,
-                        self.deduct_when,
-                        self.on_breach,
-                        self.cost,
-                    )
-                )
+                static_limits = self.static_limits
             except ValueError as e:
                 self.limiter.logger.error(
                     "failed to configure %s %s (%s)",
@@ -1123,7 +1159,7 @@ class LimitDecorator:
                 )
 
         if isinstance(obj, flask.Blueprint):
-            if dynamic_limit:
+            if not self.is_static:
                 self.limiter.limit_manager.add_runtime_blueprint_limits(
                     name, dynamic_limit
                 )
@@ -1134,20 +1170,35 @@ class LimitDecorator:
             return None
         else:
             self.limiter._marked_for_limiting.add(name)
-
-            if dynamic_limit:
-                self.limiter.limit_manager.add_runtime_route_limits(name, dynamic_limit)
+            if not self.is_static:
+                self.limiter.limit_manager.add_decorated_runtime_limit(
+                    name, dynamic_limit
+                )
             else:
-                self.limiter.limit_manager.add_static_route_limits(name, *static_limits)
+                self.limiter.limit_manager.add_decorated_static_limit(
+                    name, *static_limits
+                )
 
             @wraps(obj)
             def __inner(*a: P.args, **k: P.kwargs) -> R:
                 if (
                     self.limiter._auto_check
-                    and not self.limiter.context.rate_limiting_complete
+                    and not self.limiter.context.rate_limiting_complete.get(name, False)
                 ):
-                    self.limiter._check_request_limit(False)
-                    self.limiter.context.rate_limiting_complete = True
+                    if flask.request.endpoint:
+                        view_func = flask.current_app.view_functions.get(
+                            flask.request.endpoint, None
+                        )
+                        if view_func and not get_qualified_name(view_func) == name:
+                            self.limiter.limit_manager.add_endpoint_hint(
+                                flask.request.endpoint, name
+                            )
+
+                    self.limiter._check_request_limit(
+                        in_middleware=False, callable_name=name
+                    )
+
+                    self.limiter.context.rate_limiting_complete[name] = True
                 return cast(
                     R, flask.current_app.ensure_sync(cast(Callable[P, R], obj))(*a, **k)
                 )
